#lang racket/base

(require
 redex/reduction-semantics
 "1-lc-syntax.rkt"
 racket/contract)

(provide (all-defined-out))

;; A normalization for the lambda calculus
(define a->
  (reduction-relation eval-lcL
   #:domain e
   #:codomain e
   (--> (in-hole E (let ([x e_1]) e_2))
        (let ([x e_1]) (in-hole E e_2))
        (side-condition
         (not (redex-match? eval-lcL hole (term E))))
        "A_1")
   (--> (in-hole E (ifz v e_1 e_2))
        (ifz v (in-hole E e_1) (in-hole E e_2))
        (side-condition (not (redex-match? eval-lcL hole (term E))))
        "A_2")
   (--> (in-hole E (op v ...))
        (let ([x (op v ...)])
          (in-hole E x))
        (side-condition
         (not (redex-match? eval-lcL hole (term E))))
        (side-condition
         (not (redex-match? eval-lcL (in-hole E_1 (let ([x_1 hole]) e)) (term E))))
        (fresh x)
        "A_31")
   (--> (in-hole E (apply v ...))
        (let ([x (apply v ...)])
          (in-hole E x))
        (side-condition
         (not (redex-match? eval-lcL hole (term E))))
        (side-condition
         (not (redex-match? eval-lcL (in-hole E_1 (let ([x_1 hole]) e)) (term E))))
        (fresh x)
        "A_32")))

(define a->* (compatible-closure a-> eval-lcL e))

(define-extended-language ANFL baseL
  [M ::= N (let ([x N]) M) (ifz V M M)]
  [N ::= V (op V ...) (apply V V)]
  [V ::= x ι (lambda (x) M)]

  [F ::=  (let ([x hole]) M)]
  [K ::= mt (F :: K)])

(define/contract (a-normalize e)
  (-> (redex-match? lcL e)
      (redex-match? ANFL M))
  (car (apply-reduction-relation* a->* e)))

(module+ test
  (test-->> a->*
            intro-example
            (term
             (let ((x (apply f 5)))
               (+ 0 6))))

  (test-->> a->*
            running-example
            (term (let ((x1 (+ 2 2))) (let ((x 1)) (let ((x2 (apply f x))) (+ x1 x2))))))

  ;; Takes an annoyingly long time to run. Maybe 30 seconds on desktop.
  (test-->> a->*
            nested-branches
            (term
             (ifz 0 (ifz 0
                         (ifz 0
                              (let ([x 0]) LARGE)
                              (let ([x 1]) LARGE))
                         (ifz 1
                              (let ([x 0]) LARGE)
                              (let ([x 1]) LARGE)))
                  (ifz 1
                       (ifz 0
                            (let ([x 0]) LARGE)
                            (let ([x 1]) LARGE))
                       (ifz 1
                            (let ([x 0]) LARGE)
                            (let ([x 1]) LARGE)))))))

;; non-strict evaluation contexts
(define-extended-language meval-lcL eval-lcL
  [En ::= hole
     (ifz En e e)
     (apply v ... En e ...)
     (op v ... En e ...)])

;; Monadic form is ANF over the non-strict evaluation contexts.
;; V in operand position is deceptive, since we consider all expressions as effectful.
(define b->
  (reduction-relation meval-lcL
   #:domain e
   #:codomain e
   (--> (in-hole En (let ([x e_1]) e_2))
        (let ([x e_1]) (in-hole En e_2))
        (side-condition
         (not (redex-match? meval-lcL hole (term En))))
        "B_1")
   (--> (in-hole En (ifz v e_1 e_2))
        (let ([x (ifz v e_1 e_2)])
          (in-hole En x))
        (side-condition
         (not (redex-match? meval-lcL hole (term En))))
        (fresh x)
        "B_2'")
   (--> (in-hole En (op v ...))
        (let ([x (op v ...)])
          (in-hole En x))
        (side-condition
         (not (redex-match? meval-lcL hole (term En))))
        (fresh x)
        "B_31")
   (--> (in-hole En (apply v ...))
        (let ([x (apply v ...)])
          (in-hole En x))
        (side-condition
         (not (redex-match? meval-lcL hole (term En))))
        (fresh x)
        "B_32")))

(define b->* (compatible-closure b-> meval-lcL e))

(define-extended-language monadicL baseL
  [C ::= U (let ([x C]) C) (ifz U C C) (op U ...) (apply U U)]
  [U ::= x ι (lambda (x) C)]

  [F ::=  (let ([x hole]) M)]
  [K ::= mt (F :: K)])

(define (b-normalize e)
  (car (apply-reduction-relation* b->* e)))

(module+ test
  (test-->> b->*
            intro-example
            (term
             (let ((x (apply f 5)))
               (+ 0 6))))

  (test-->> b->*
            running-example
            (term (let ((x1 (+ 2 2))) (let ((x 1)) (let ((x2 (apply f x))) (+ x1 x2))))))

  (test-->> b->*
           nested-branches
           (term
            (let ([x (let ([x1 (ifz 0 0 1)])
                       (let ([x2 (ifz x1 0 1)])
                         (ifz x2
                              0
                              1)))])
              LARGE))))
