#lang racket/base

(require
 redex/reduction-semantics
 "1-lc-syntax.rkt"
 racket/contract)
(provide (all-defined-out))

(define-extended-language imonadicL baseL
  [P ::= t]
  [t ::= (begin s ... t) v (ifz v t t) (call v v) (op v ...)]
  [s ::= (begin s ...) (ifz v s s) (set! x v) (set! x t) (set! x (op v ...))
     (set! x (call v v))]
  [v ::= (lambda (x) t) ι x]

  ;; Compatible closure of S/T
  [C ::=
     ;(compatible-closure-context t)
     ;(compatible-closure-context s)
     (compatible-closure-context t #:wrt s)
     (compatible-closure-context s #:wrt t)]
  #:binding-forms
  (lambda (x) t #:refers-to x))

(define-extended-language abnfL baseL
  [P ::= t]
  [t ::= (begin s ... t) v (ifz v t t) (call v v) (op v ...)]
  [s ::= (begin s ...) (ifz v s s) (set! x v) (set! x (op v ...)) (set! x (call v v))]
  [v ::= (lambda (x) t) ι x]
  #:binding-forms
  (lambda (x) t #:refers-to x))

(define ab->
  (reduction-relation
   imonadicL
   #:domain s
   #:codomain s

   (--> (set! x (ifz v t_1 t_2))
        (ifz v (set! x t_1) (set! x t_2)))

   (--> (set! x (begin s ... t))
        (begin s ... (set! x t)))))

(define ab->*
  (context-closure ab-> imonadicL C))

(define/contract (ab-normalize e)
  (-> (redex-match? imonadicL t)
      (redex-match? abnfL t))
  (car (apply-reduction-relation* ab->* e)))

(module+ test
  ;; Equivalent to (code-gen (term (let ((x1 (+ 2 2))) (let ((x 1)) (let ((x2 (apply f x))) (+ x1 x2))))))
  (test-match imonadicL P (term
                           (begin
                             (set! x1 (+ 2 2))
                             (set! x 1)
                             (set! x2 (call f x))
                             (+ x x))))
  (test-->>
   ab->*
   (term
    (begin
      (set! x1 (+ 2 2))
      (set! x 1)
      (set! x2 (call f x))
      (+ x x)))
   (term
    (begin
      (set! x1 (+ 2 2))
      (set! x 1)
      (set! x2 (call f x))
      (+ x x))))


  #;(let ([x (let ([x1 (ifz 0 0 1)])
             (let ([x2 (ifz x1 0 1)])
               (ifz x2
                    0
                    1)))])
    LARGE)

  (test-->>
   ab->*
   (term
    (begin
      (set! x
            (begin
              (set! x1 (ifz 0 0 1))
              (set! x2 (ifz x1 0 1))
              (ifz x2
                   0
                   1)))
      LARGE))
   (term
    (begin
      (begin
        (ifz 0 (set! x1 0) (set! x1 1))
        (ifz x1 (set! x2 0) (set! x2 1))
        (ifz x2
             (set! x 0)
             (set! x 1)))
      LARGE))))


;; TODO: CEK Machine
(define-extended-language cek-imonadicL imonadicL
  [F ::= (begin (set! x hole) s ... t)]
  [K ::= mt (F :: K)]
  [wv ::= ι x]
  [hv ::= ι (closure Σ (lambda (x) t))]
  [Σ ::= () (Σ (x hv))])

(define-metafunction cek-imonadicL
  deref : Σ x -> hv
  [(deref (Σ (x hv)) x)
   hv]
  [(deref (Σ (x_!_! hv)) (name x x_!_1))
   (deref Σ x)])

(define-metafunction cek-imonadicL
  resolve-wv : Σ wv -> hv
  [(resolve-wv Σ ι) ι]
  [(resolve-wv Σ x)
   (deref Σ x)])

(define cek->
  (reduction-relation
   cek-imonadicL
   #:domain (t Σ K)
   #:codomain (t Σ K)

   (--> ((begin (set! x wv) s ... t) Σ K)
        ((begin s ... t) (Σ (x (resolve-wv Σ wv))) K)
        "move")

   ;; assume for simplicity that all lambdas are compiled as effects into this machine
   ;; easy to normalize this during B reduction by treating λ as an effect...
   ;; which, it kind of is.
   (--> ((begin (set! x (lambda (x) t)) s ... t) Σ K)
        ((begin s ... t) (Σ (x (closure Σ (lambda (x) t)))) K)
        "closure")

   (--> ((begin (set! x (call x_1 wv_2)) s ... t) Σ K)
        (t_1 (Σ_1 (y wv_2))
         ((begin (set! x hole) s ... t) :: K))
        (where (closure Σ_1 (lambda (y) t_1))
               (deref Σ x))
        "call")

   (--> (v Σ (F :: K))
        ((in-hole F v) Σ K)
        "return")

   (--> ((begin (ifz wv s_1 s_2) s ... t) Σ K)
        ((begin s_1 s ... t) Σ K)
        (where 0 (resolve-wv Σ wv))
        "ifz")

   (--> ((begin (ifz wv s_1 s_2) s ... t) Σ K)
        ((begin s_2 s ... t) Σ K)
        (side-condition (not (equal? (term (resolve-wv Σ wv)) 0)))
        "ifnz")

   (--> ((begin (set! x (op wv ...)) s ... t) Σ K)
        ((begin s ... t) (Σ (x (δ op (resolve-wv wv) ...))) K)
        "primop")

   (--> ((op wv ...) Σ K)
        ((δ op (resolve-wv wv) ...) Σ K)
        "tail-primop")

   (--> ((ifz wv t_1 t_2) Σ K)
        (t_1 Σ K)
        (where 0 (resolve-wv Σ wv))
        "ifz-t")

   (--> ((ifz wv t_1 t_2) Σ K)
        (t_2 Σ K)
        (side-condition (not (equal? (term (resolve-wv Σ wv)) 0)))
        "ifnz-t")

   (--> ((begin (set! x t_1) s ... t_2) Σ K)
        (t_1 Σ ((begin (set! x hole) s ... t_2) :: K))
        (side-condition
         (not (redex-match? cek-imonadicL wv (term t_1))))
        (side-condition
         (not (redex-match? cek-imonadicL (op wv ...) (term t_1))))
        (side-condition
         (not (redex-match? cek-imonadicL (call wv ...) (term t_1))))
        "ab")

   (--> ((begin (begin s_1 ...) s_2 ... t_2) Σ K)
        ((begin s_1 ... s_2 ... t_2) Σ K)
        "admin-1")

   (--> ((begin t) Σ K)
        (t Σ K)
        "admin-2")))

(define (init-cek term)
  `(,term () mt))

(require racket/match)
(require (only-in "3-machine-semantics.rkt"
                  stack-length))

(define (cek-max-stack states)
  (apply max (map cek-stack-length states)))

(define (cek-stack-length state)
  (match state
    [`(,c ,E ,K)
     (stack-length K)]))

(module+ test
  (require rackunit)
  (define-term monadic-large-eg
    (begin
      (set! x
            (begin
              (set! x1 (ifz 0 0 1))
              (set! x2 (ifz x1 0 1))
              (ifz x2
                   0
                   1)))
      LARGE))

  (define-term abnf-large-eg
    (begin
      (begin
        (ifz 0 (set! x1 0) (set! x1 1))
        (ifz x1 (set! x2 0) (set! x2 1))
        (ifz x2
             (set! x 0)
             (set! x 1)))
      LARGE))

  ;; AB-normalization optimizes the stack, without duplicating the continuation or join points.
  (check-equal?
   (cek-max-stack (apply-reduction-relation* #:all? #t cek-> (term (monadic-large-eg () mt))))
   2)

  (check-equal?
   (cek-max-stack (apply-reduction-relation* #:all? #t cek-> (term (abnf-large-eg () mt))))
   0))

(define-extended-language rimonadicL baseL
  [P ::= t]
  [t ::= (begin s ... t) rv (alloc r v) (ifz rv t t) (call rv rv) (alloc r (op rv ...))]
  [s ::= (begin s ...)
     (ifz rv s s)
     (set! x (alloc r v))
     (set! x rv)
     (set! x t)
     (set! x (alloc r (op rv ...)))
     (set! x (call rv rv))
     (ralloc r)
     (rfree r)]
  [v ::= ι x (lambda (x) t) a]
  [r ::= (variable-prefix r)]
  [x ::= variable-not-otherwise-mentioned]
  [o ::= (variable-prefix o)]
  [hv ::= (closure Σ (lambda (x) t)) ι a]
  [wv ::= ι a]
  [rv ::= a x]
  [a ::= (r o)]

  [F ::= (begin (set! x hole) s ... t)]
  [K ::= mt (F :: K)]
  [H ::= () (H (o hv))]
  [S ::= () (S (r H))]
  [Σ ::= () (Σ (x a))]

  ;; Compatible closure of S/T, for ab-normalization
  [C ::=
     (compatible-closure-context t)
     (compatible-closure-context s)
     (compatible-closure-context t #:wrt s)
     (compatible-closure-context s #:wrt t)]
  )


;; Duplicated because metafunctions aren't scoped to languages
(define-metafunction rimonadicL
  do-load : S a -> hv
  [(do-load (S (r H)) (r o))
   (do-H-load H o)]
  [(do-load (S (r_!_! H)) ((name r r_!_1) o))
   (do-load S (r o))])

(define-metafunction rimonadicL
  do-H-load : H o -> hv
  [(do-H-load (H (o hv)) o)
   hv]
  [(do-H-load (H (o_!_1 hv)) (name o o_!_1))
   (do-H-load H o)])


(define-metafunction rimonadicL
  do-rfree : S r -> S
  [(do-rfree (S (r H)) r)
   S]
  [(do-rfree (S ((name r r_!_1) H)) (name r_2 r_!_1))
   ((do-rfree S r_2) (r H))])

(define-metafunction rimonadicL
  resolve-rv : Σ rv -> hv
  [(resolve-rv Σ a) a]
  [(resolve-rv Σ x)
   (do-lookup Σ x)])

(define-metafunction rimonadicL
  do-lookup : Σ x -> a
  [(do-lookup (Σ (x a)) x)
   a]
  [(do-lookup (Σ (x_!_! a)) (name x x_!_1))
   (do-lookup Σ x)])

(define-metafunction rimonadicL
  do-alloc : S r hv -> (S a)
  [(do-alloc S r hv)
   ((insert S (r o) hv) (r o))
   (where o ,(gensym 'o))])

(define-metafunction rimonadicL
  insert : S a hv -> S
  [(insert (S (r H)) (r o) hv)
   (S (r (H (o hv))))]
  [(insert (S (r_1 H)) a hv)
   ((insert S a hv) (r_1 H))])

(define cesk->
  (reduction-relation
   rimonadicL
   #:domain (t Σ S K)
   #:codomain (t Σ S K)

   (--> ((begin (set! x rv) s ... t) Σ S K)
        ((begin s ... t) (Σ (x (resolve-rv Σ rv))) S K)
        "move")

   (--> ((begin (set! x (alloc r wv)) s ... t) Σ S K)
        ((begin s ... t) (Σ (x a)) S_1 K)
        (where (S_1 a) (do-alloc S r wv))
        "alloc")

   (--> ((begin (set! x (alloc r (lambda (x_1) t_1))) s ... t) Σ S K)
        ((begin s ... t) (Σ (x a)) S_1 K)
        (where (S_1 a) (do-alloc S r (closure Σ (lambda (x_1) t_1))))
        "closure-alloc")

   (--> ((begin (ralloc r) s ... t) Σ S K)
        ((begin s ... t) Σ (S (r ())) K)
        "ralloc")

   (--> ((begin (rfree r) s ... t) Σ S K)
        ((begin s ... t) Σ (do-rfree S r) K)
        "rfree")

   (--> ((begin (set! x (call rv_1 rv_2)) s ... t) Σ S K)
        (t_1 (Σ_1 (x_1 a_2)) S
         ((begin (set! x hole) s ... t) :: K))
        (where a_1 (resolve-rv Σ rv_1))
        (where a_2 (resolve-rv Σ rv_2))
        (where (closure Σ_1 (lambda (x_1) t_1))
               (do-load S a_1))
        "call")

   (--> (rv Σ S (F :: K))
        ((in-hole F a) Σ S K)
        (where a (resolve-rv Σ rv))
        "return")
   (--> ((call rv_1 rv_2) Σ S K)
        (t_1 (Σ_1 (x_1 a_2)) S K)
        (where a_1 (resolve-rv Σ rv_1))
        (where a_2 (resolve-rv Σ rv_2))
        (where (closure Σ_1 (lambda (x_1) t_1))
               (do-load S a_1))
        "tail-call")

   (--> ((begin (ifz rv s_1 s_2) s ... t) Σ S K)
        ((begin s_1 s ... t) Σ S K)
        (where a (resolve-rv Σ rv))
        (where 0 (do-load S a))
        "ifz")

   (--> ((begin (ifz rv s_1 s_2) s ... t) Σ S K)
        ((begin s_2 s ... t) Σ S K)
        (where a (resolve-rv Σ rv))
        (side-condition (not (equal? (term (do-load S a)) 0)))
        "ifnz")

   (--> ((begin (set! x (alloc r (op rv ...))) s ... t) Σ S K)
        ((begin s ... t) (Σ (x a_r)) S_r K)
        (where (a ...) ((resolve-rv Σ rv) ...))
        (where (hv ...) ((do-load S a) ...))
        (where hv_r (δ op hv ...))
        (where (S_r a_r) (do-alloc S r hv_r))
        "primop")

   (--> ((alloc r (op rv ...)) Σ S K)
        (a_r Σ S_r K)
        (where (a ...) ((resolve-rv Σ rv) ...))
        (where (hv ...) ((do-load S a) ...))
        (where hv_r (δ op hv ...))
        (where (S_r a_r) (do-alloc S r hv_r))
        "tail-primop")

   (--> ((ifz rv t_1 t_2) Σ S K)
        (t_1 Σ S K)
        (where a (resolve-rv Σ rv))
        (where 0 (do-load S a))
        "ifz-t")

   (--> ((ifz rv t_1 t_2) Σ S K)
        (t_2 Σ S K)
        (where a (resolve-rv Σ rv))
        (side-condition (not (equal? (term (do-load S a)) 0)))
        "ifnz-t")

   (--> ((begin (set! x t_1) s ... t_2) Σ S K)
        (t_1 Σ S ((begin (set! x hole) s ... t_2) :: K))
        (side-condition
         (not (redex-match? rimonadicL (alloc r any) (term t_1))))
        (side-condition
         (not (redex-match? rimonadicL (call rv_1 rv_2) (term t_1))))
        (side-condition
         (not (redex-match? rimonadicL rv (term t_1))))
        "ab")

   (--> ((begin (begin s_1 ...) s_2 ... t_2) Σ S K)
        ((begin s_1 ... s_2 ... t_2) Σ S K)
        "admin-1")

   (--> ((begin t) Σ S K)
        (t Σ S K)
        "admin-2")))

(define (init-cesk term)
  `(,term () (() (r0 ())) mt))


;; NOTE: copypaste from 3-machine-semantics.rkt

;; Read back the final address out of the machine state.
;; There must be exactly 1 final state.
(define (cesk-read-val states)
  (match states
    [`((,rv ,Σ ,S mt))
     (term (do-load ,S (resolve-rv ,Σ ,rv)))]
    [_ (error "terminated at invalid state" states)]))

;; Hand compilation of
#|
(letregion r4
           (apply
            (let ([x (@ r1 2)])
              (@ r4 (lambda (y) (@ r1 (* x y)))))
            (@ r1 4)))
|#

(define rimonadic-eg1-term
  (term
   (begin
     (ralloc r4)
     (set! x (alloc r0 2))
     (set! x1 (alloc r4 (lambda (y)
                          (alloc r0 (* x y)))))
     (set! x2 (alloc r0 4))
     (set! x3 (call x1 x2))
     (rfree r4)
     x3)))

(module+ test
  (define (final-state-equiv? state expected-val)
    (equal? (cesk-read-val (list state)) expected-val))

  (test-match rimonadicL
              t
              rimonadic-eg1-term)

  (test-->> cesk->
            #:equiv final-state-equiv?
            (init-cesk rimonadic-eg1-term)
            8))

;; enough to get examples working
(define (code-gen-region e)
  (match e
    [`(letregion ,r ,e)
     (define x (gensym 'x))
     `(begin
        (ralloc ,r)
        (set! ,x ,(code-gen-region e))
        (rfree ,r)
        ,x)]
    [`(let (,x ,e1) ,e2)
     `(begin
        (set! ,x ,(code-gen-region e1))
        ,(code-gen-region e2))]
    [`(@ ,r ,v)
     `(alloc ,r ,(code-gen-region v))]
    [`(,op ,vs ...)
     #:when (op? op)
     `(,op ,@(map code-gen-region vs))]
    [(? value?)
     e]))

(define lst:code-gen-anfregion-eg1
  (code-gen-region
   '(letregion r2
               (letregion r1
                          (let (x (@ r1 1))
                            (let (x1 (@ r1 2))
                              (let (x2 (@ r2 (* x x1)))
                                (letregion r3
                                           (let (x3 (@ r3 3))
                                             (let (x4 (@ r3 4))
                                               (let (x5 (@ r2 (* x3 x4)))
                                                 (@ r0 (* x2 x5)))))))))))))

(define lst:code-gen-mregion-eg1
  (code-gen-region
   '(letregion r2
               (let (x4 (letregion r1
                                   (let (x2 (@ r1 1))
                                     (let (x3 (@ r1 2))
                                       (@ r2 (* x2 x3))))))
                 (let (x5 (letregion r3
                                     (let (x (@ r3 3))
                                       (let (x1 (@ r3 4))
                                         (@ r2 (* x x1))))))
                   (@ r0 (* x4 x5)))))))



;; Extend ab normalization to rimonadic
(define rab->
  (reduction-relation
   rimonadicL
   #:domain s
   #:codomain s

   (--> (set! x (ifz v t_1 t_2))
        (ifz v (set! x t_1) (set! x t_2)))

   (--> (set! x (begin s ... t))
        (begin s ... (set! x t)))))

(define rab->*
  (context-closure rab-> rimonadicL C))

(define (rab-normalize e)
  (car (apply-reduction-relation* rab->* e)))

(define lst:rabnf-mregion-eg1
  (rab-normalize lst:code-gen-mregion-eg1))

;; For pretty printing
;; (define flatten-begin->
;;   (reduction-relation
;;    rimonadicL
;;    #:domain t
;;    #:codomain t
;;    (--> (begin s_1 ... (begin s_2 ...) s_3 ... t)
;;         (begin s_1 ... s_2 ... s_3 ... t))))
;;
;; (define flatten-begin->*
;;   (compatible-closure flatten-begin-> rimonadicL s))
;;
;; (define (flatten-begin e)
;;   (car (apply-reduction-relation* flatten-begin->* e)))

;; A language-generic version
(require (only-in racket/list append-map))
(define (flatten-begin t)
  `(begin
     ,@(let loop ([t t])
         (match t
           [`(begin ,s ...)
            (append-map loop s)]
           [n (list n)]))))

;; NOTE Copypasta from machine-semantics.rkt

(require (only-in "3-machine-semantics.rkt"
                  snoc-list-len
                  store-count-addrs
                  stack-length))

;; Count the maximum number of regions in the machine state
(define (cesk-max-regions states)
  (apply max (map cesk-regions-size states)))

(define (cesk-regions-size state)
  (match state
    [`(,c ,E ,S ,K)
     (snoc-list-len S)]))

;; Compute the maximum number of unfreed addresses in the machine state.
(define (cesk-max-memory states)
  (apply max (map cesk-memory-size states)))

(define (cesk-memory-size state)
  (match state
    [`(,c ,E ,S ,K)
     (store-count-addrs S)]))

(define (cesk-max-stack states)
  (apply max (map cesk-stack-length states)))

(define (cesk-stack-length state)
  (match state
    [`(,c ,E ,S ,K)
     (stack-length K)]))

(module+ test
  ;; Monadic then AB-normalization does increase max regions or memory
  (let ([D_A (apply-reduction-relation* #:all? #t cesk->
                                        ;; Must re AB-normalize, since our code
                                        ;; generator is for monadic and not ANF.
                                        (init-cesk (rab-normalize lst:code-gen-anfregion-eg1)))]
        [D_B (apply-reduction-relation* #:all? #t cesk->
                                        (init-cesk (rab-normalize lst:code-gen-mregion-eg1)))])
    (check
     >
     (cesk-max-regions D_A)
     (cesk-max-regions D_B))

    (check
     >
     (cesk-max-memory D_A)
     (cesk-max-memory D_B))

    ;; But does not increase cesk-max-stack
    (check-equal?
     (cesk-max-stack D_A)
     (cesk-max-stack D_B))))

;; Examples for paper
(module+ main
  (flatten-begin (rab-normalize lst:code-gen-anfregion-eg1))
  (flatten-begin (rab-normalize lst:code-gen-mregion-eg1)))
